#include <cmath>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include "GMesh.hpp"
#include "error.hpp"
#include "throwassert.hpp"

extern long verbosity ;

namespace GMesh3D  
{
void Mesh::read(const char * maillage)
{
   int i,i0,i1,i2,i3,ir;   
   volume=0;
   
   // Existence du maillage
   ifstream f(maillage);
   if (!f) 
   {  cerr << "Erreur ouverture du fichier " << maillage << endl;
      throw(ErrorExec("exit",1));
   }
   if(verbosity)
     cout << "Lecture du fichier maillage\"" <<maillage<<"\""<<  endl;
     
   char lec[50];
   
   // Recherche de la dimension
   f >> lec;
   while(strcmp(lec,"Dimension")!=0)
   {
      f>>lec;
   }
   f>>dim;
   if(verbosity)
     cout << "   Dimension du maillage : " << dim << endl;

   // Recherche des sommets
   f >> lec;
   while(strcmp(lec,"Vertices")!=0)
   {
      f>>lec;     
   }
   f>>nv;
   if(verbosity)
     cout << "   Nb de sommet          : " <<nv<< endl;
   vertices   = new Vertex[nv];   
   // lecture des coordonnees
   for (i=0;i<nv;i++) 
   {  
     f >> vertices[i],throwassert(f.good());   
     vertices[i].color_efface();
   }  
   
   // Recherche des Tetraedres
   f >> lec;
   while(strcmp(lec,"Tetrahedra")!=0)
   {
      f>>lec;
   }
   f>>nt;   
   if(verbosity)
     cout << "   Nb de Tetraedres      : " <<nt<<  endl;
   tetraedres = new Tetraedre [nt];
   // lecture des 4 numeros de sommets des tetraedres et le label
   for (i=0;i<nt;i++) 
   {
     f >> i0 >> i1 >> i2 >>i3 >> ir;
     throwassert(f.good() && i0>0 && i0<=nv && i1>0 && i1<=nv && i2>0 && i2<=nv&& i3>0 && i3<=nv);
     tetraedres[i].init(vertices,i0-1,i1-1,i2-1,i3-1,ir,0); 
     volume += tetraedres[i].volume;
     tetraedres[i].color_efface();}    
   f.close();       
}


R Mesh::taille_caracteristique(void)
{
	R h=0.;
	for (int t=0;t<nt;t++) {
		Tetraedre &K = tetraedres[t];
		for (int j=0;j<3;j++) {
			R3 A =*(K.vertices[j]);
			for(int i =j+1; i<4;i++) {
				R3 B =*(K.vertices[i]);
				h = Max(h,Norme2_2(B-A));
			}
		}
	}
	return h;
}

void Mesh::Buildadjacence()
{
    // Tableau tmp
    int *Linktmp, *Mintmp, *Maxtmp;
    int *Sumtmp;
    int *indicetmp; 
    int *tettmp;
    
    int nsmax = 4*nt;   

    cout << "   Buildadjacence...       ";
    cout.flush();

    Linktmp   = new int [nsmax];
    Mintmp    = new int [nsmax]; 
    Maxtmp    = new int [nsmax];
    indicetmp = new int [nsmax];
    tettmp = new int [nsmax];

    Sumtmp    = new int [3*nv];

    // initialisation des tableaux a -1
    for (int i=0;i<nsmax;i++)
    {
       Linktmp[i]   = -1;
       Mintmp[i]    = -1;
       Maxtmp[i]    = -1;
       indicetmp[i] = -1;
       tettmp[i]    = -1;
    }
    for (int i=0;i<(3*nv);i++)
       Sumtmp[i] = -1;
 

    int ns = 0;
    for(int i=0;i<nt;i++)
    {
       for(int j=0;j<4;j++)
       {
          // trois sommet = une face
          int s1 = number(tetraedres[i].vertices[j]);
          int s2 = number(tetraedres[i].vertices[(j+1)%4]);
          int s3 = number(tetraedres[i].vertices[(j+2)%4]);
          int s  = s1 + s2 + s3;
          int l=Sumtmp[s];
          if (l == -1) 
          {
              Sumtmp[s]     = ns;
              Mintmp[ns]    = Min(s1,s2,s3);
              Maxtmp[ns]    = Max(s1,s2,s3);
              indicetmp[ns] = (j+3)%4;
              tettmp[ns]    = i;            
              ns++;
          }
          else
          {
              int jecherche = -1;	   
              while(jecherche == -1)
              {	           
                 if (Mintmp[l]==Min(s1,s2,s3) && Maxtmp[l]==Max(s1,s2,s3))
                 {
                     // On a trouve la meme face
                     tetraedres[i].voisinage(tettmp[l],(j+3)%4); 
                     tetraedres[tettmp[l]].voisinage(i,indicetmp[l]); 
                     jecherche = 1;
                 }    
                 else
                 {
                    if (Linktmp[l] == -1)  
                    {
                        // Nouvelle face que l'on insere 
                        Linktmp[l] = ns;
                        Mintmp[ns]    = Min(s1,s2,s3);
                        Maxtmp[ns]    = Max(s1,s2,s3);
                        indicetmp[ns] = (j+3)%4;
                        tettmp[ns]    = i;
                        ns++;
                        jecherche = 1;
                    }
                    else
                    {
		       l= Linktmp[l];
                    } 
                 }             
              }
          }
       }
    }
    delete [] Linktmp;
    delete [] Mintmp;
    delete [] Maxtmp;
    delete [] indicetmp;
    delete [] tettmp; 
    delete [] Sumtmp;   

    cout << "finished" << endl; 
}

Mesh::~Mesh()
 {
   delete [] tetraedres;
   delete [] vertices;
 }
 
void Result::read(const char * solution, Mesh & Th)
{
   scalaires = 0;
   velocity = 0;

   // Existence de la solution
   ifstream f(solution);
   if (!f) 
   {  cerr << "Erreur ouverture du fichier " << solution << endl;
      throw(ErrorExec("exit",1));
   }
   if(verbosity)
     cout << "Lecture du fichier solution \"" <<solution<<"\""<<  endl;

   // Lecture de la dimension, type de solution, nbre de sommet, jenesaispas
   f >> dim >> type >> nv >> typedonnees;
   
   if (dim !=Th.dim)
   {
      cerr << "Erreur dans le fichier les dimensions ne sont pas les memes " << endl;
      throw(ErrorExec("exit",1));
   }
   
   if (nv !=Th.nv)
   {
      cerr << "Erreur dans le fichier : il n'y a pas le meme nombre de sommets " << endl;
      throw(ErrorExec("exit",1));
   }   
   
   if(verbosity)
     cout << "   Dimension du maillage : " << dim << endl;     

   switch(type) {
	   default:
      		cerr << "Type inconnu "<< type << ", on ne sait pas traiter " << endl;
		throw(ErrorExec("exit",1)); 	  
		break;
           case RESULT_TYPE_SCALAIRE:
          	cout << "   Type de solution      : scalaire" << endl;
                scalaires = new R[nv];
                for (int i=0;i<nv;i++)
                    f >> scalaires[i];
		break;
           case RESULT_TYPE_VECTORIEL:
		cout << "   Type de solution      : vitesse" << endl;	  
                velocity = new R3[nv];
                for (int i=0;i<nv;i++)
                    f >> velocity[i];
		break;
   };

   if(verbosity)
     cout << "   Nb de sommet          : " <<nv<< endl;   

   f.close();       
} 

Result::Result(Mesh & Th2,int letype, int letypedonnees )
{
   scalaires = 0;
   velocity = 0;

   // Cette fonction va nous permettre de preparer le terrain pour la solution interpolee 
   dim  = Th2.dim;
   type = letype;
   nv   = Th2.nv;  
   typedonnees = letypedonnees;
   
   // Allocation dynamique des tableaux
   if ( RESULT_TYPE_SCALAIRE == type)
   { 
      scalaires = new R[nv]; 
      for (int i=0;i<nv;i++)
	      scalaires[i] = 0.;
   }   
   if ( RESULT_TYPE_VECTORIEL == type)
   {
      velocity = new R3[nv];
      for (int i=0;i<nv;i++)
      { velocity[i].x = 0.; velocity[i].y = 0.; velocity[i].z = 0.;}
   }	 
} 
 
}  // namespace GMesh3D  

