//-------------------------------------------------------------------
// Creation du code par Hanen AMOR  & Thomas CAPRICELLI
//-------------------------------------------------------------------
#include <stdio.h>
#include <ctime>
#include <cmath>
#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>       
#include <string> 
#include <set>

#include "GMesh.hpp" 
#include "error.hpp"
#include "throwassert.hpp"

#define VERSION_STRING "Projet 4 Hecht/Frey, version 0.9, juin 2004"

using namespace std;
using namespace GMesh3D;
int verbosity =2;
int TheCurrentLine=0;
double  Epsilon = 1e-4;
int nb_tetraedre_parcourus = 0;

#define FREQU_AFFICHE 250234325

R champs_scalaire_lineaire( R3 & P);
void test(const char * maillage1,  const char *maillage2);

// fonction qui calcule la coord barycentrique
R coord_bary(R3 & P, R3 &A,R3 &B, R3 &C,R3 &D)
{
   R3     temp =  (B-C)^(B-D);
   R coord = (temp,(B-P))/(temp,(B-A));
   return (coord);
}


// fonction qui calcule la coord barycentrique pour une projection2d
R coord_bary2d(R3 & P, R3 &A,R3 &B, R3 &C)
{
   R3 V   = ((B-A)^(C-A))^(C-B);  
   R coord = (P-B,V)/(A-B,V);
   return (coord);
}

// Chemin de promenade dans le maillage
// retourne l'indice du tetraedre qui contient le
// point, ou -1 si on a pas trouve
// arguments : 
// 	on utilise le Mesh Th,
// 	on Cherche le tetraedre contenant P
// 	et on part du tetradre Th[t]
int promenade(Mesh & Th, R3 &P, int t)
{    
    Tetraedre *T;
    while ( t!=-1 ) 
    {

        int kk = 0;
        int dd = 0;
        int tetadj[4];
        int tetadjhorsvide[4];

    	Tetraedre *T= &Th[t];
        R3 &A =*(T->vertices[0]);    R3 &B =*(T->vertices[1]);
        R3 &C =*(T->vertices[2]);    R3 &D =*(T->vertices[3]);            

        // on regarde quelles sont les coo barycentriques positives
        if(coord_bary(P,A,B,C,D) < -Epsilon) 
          { tetadj[kk++] = T->Neigh[0]; if (T->Neigh[0] != -1) {tetadjhorsvide[dd++] = T->Neigh[0];  } }
        if(coord_bary(P,B,C,D,A) < -Epsilon) 
          { tetadj[kk++] = T->Neigh[1]; if (T->Neigh[1] != -1) {tetadjhorsvide[dd++] = T->Neigh[1];  } }
        if(coord_bary(P,C,D,A,B) < -Epsilon) 
          { tetadj[kk++] = T->Neigh[2]; if (T->Neigh[2] != -1) {tetadjhorsvide[dd++] = T->Neigh[2];  } }
        if(coord_bary(P,D,A,B,C) < -Epsilon) 
          { tetadj[kk++] = T->Neigh[3]; if (T->Neigh[3] != -1) {tetadjhorsvide[dd++] = T->Neigh[3];  } }
        
        // si elles sont toutes les quatre positives : on a trouve le tetraedre
        if(kk==0) 
            return t;	    	
        // Si on est la c'est que l'on a pas trouver et on va partir dans la direction des 
        // coord negatives a condition que dans cette direction il y un voisin 
        if (dd ==0)       
           return (-1);

	// on va au tetraedre suivant dans la 'bonne' direction (teta negatif)
	nb_tetraedre_parcourus++; 
        t = tetadjhorsvide[random()%dd]; 	
//	printf("promenade switching to %d\n", t);
    } 

    // si on est la c'est qu'il y a des problemes et on ne trouve pas le tetraedre
    return t;
}


// La fonction interpolation pour les scalaires
void interpolation_scalaires(Result & Reinter,Result & Re,int sommet, R3 &P, Mesh & Th, int t)
{
    Tetraedre & K(Th[t]);    
    assert(Reinter.scalaires[sommet] == 0. );
    for (int i=0;i<4;i++)
    { 
        R3 &A =*(K.vertices[i]);         R3 &B =*(K.vertices[(i+1)%4]);
        R3 &C =*(K.vertices[(i+2)%4]);   R3 &D =*(K.vertices[(i+3)%4]);    
        R lambda = coord_bary(P,A,B,C,D);    
        Reinter.scalaires[sommet] += Re.scalaires[Th(t,i)] *lambda;
    }
}

// La fonction interpolation pour les vitesses
void interpolation_velocity(Result & Reinter,Result & Re,int sommet, R3 &P, Mesh & Th, int t)
{
    Tetraedre & K(Th[t]);    
    assert(Reinter.velocity[sommet].x == 0.);
    assert(Reinter.velocity[sommet].y == 0.);
    assert(Reinter.velocity[sommet].z == 0.);
    for (int i=0;i<4;i++)
    { 
        R3 &A =*(K.vertices[i]);         R3 &B =*(K.vertices[(i+1)%4]);
        R3 &C =*(K.vertices[(i+2)%4]);   R3 &D =*(K.vertices[(i+3)%4]);        
        R lambda = coord_bary(P,A,B,C,D);    
        Reinter.velocity[sommet] += Re.velocity[Th(t,i)] *lambda;
    }
}

// La fonction suite interpolation pour les scalaires
int nouvelle_interpolation_scalaires(Result & Reint,Mesh & Thint,int sommet, int t)
{

    int lesvertex[3];
    int kk=0;
    int somi;

    somi = Thint(t,0);    
    if( Thint.vertices[somi].color==1){ lesvertex[kk++] = somi; }
    somi = Thint(t,1);    
    if( Thint.vertices[somi].color==1){ lesvertex[kk++] = somi; }
    somi = Thint(t,2);    
    if( Thint.vertices[somi].color==1){ lesvertex[kk++] = somi; }
    somi = Thint(t,3);    
    if( Thint.vertices[somi].color==1){ lesvertex[kk++] = somi; }
    
    if(kk!=3)
    { return(0);}
    else
    {
       // je calcule la projection du sommet !!
       R3 P = Thint(sommet);
       for(int i=0;i<3;i++)
       {  R3 &A = Thint(lesvertex[i]);
          R3 &B = Thint(lesvertex[(i+1)%3]);
          R3 &C = Thint(lesvertex[(i+2)%3]);
          R lambda = coord_bary2d(P,A,B,C);
          Reint.scalaires[sommet] += Reint.scalaires[lesvertex[i]] *lambda;
       }
      return(1);
    }
}


// La fonction suite interpolation pour les velocity
int nouvelle_interpolation_velocity(Result & Reint,Mesh & Thint,int sommet, int t)
{

    int lesvertex[3];
    int kk=0;
    int somi;

    somi = Thint(t,0);    
    if( Thint.vertices[somi].color==1){ lesvertex[kk++] = somi; }
    somi = Thint(t,1);    
    if( Thint.vertices[somi].color==1){ lesvertex[kk++] = somi; }
    somi = Thint(t,2);    
    if( Thint.vertices[somi].color==1){ lesvertex[kk++] = somi; }
    somi = Thint(t,3);    
    if( Thint.vertices[somi].color==1){ lesvertex[kk++] = somi; }
    
    if(kk!=3)
    { return(0);}
    else
    {  // je calcule la projection du sommet !!
       R3 P = Thint(sommet);
       for(int i=0;i<3;i++)
       {  R3 &A = Thint(lesvertex[i]);
          R3 &B = Thint(lesvertex[(i+1)%3]);
          R3 &C = Thint(lesvertex[(i+2)%3]);
          R lambda = coord_bary2d(P,A,B,C);
          Reint.velocity[sommet] += Reint.velocity[lesvertex[i]] *lambda;
       }
      return(1);
    }
}

/*
 * calcule pour chaque point du triangle, le tetraedre du premier
 * qui correspond. On stock ces numero dans lesTh ( taille Thint.nv.. )
 */
void  calcule_pile(Mesh &Thint, Mesh &Th, int *lesTh)
{
    // The optimisation "pile"
    Pile PTh(Thint.nt);
    Pile PThint(Thint.nt);
    PTh.push(0);          
    PThint.push(0);       
    // Inutile de revenir sur un tetraedre ou je suis deja passee    
    Thint.tetraedres[PThint.top()].colorise();    
    int nb_tetraedre_parcourus_total = 0;
    for(int i=0;i<Thint.nv;i++) lesTh[i] = -1;   

    while (PThint.kpile()>0)
    {
        // je depile
        int t      = PTh.pop();
        int tt     = t;	
        int tThint = PThint.pop();
        Tetraedre  & K(Thint[tThint]);
        // je traite les 4 sommets si pas deja fait
        for (int i=0;i<4;i++) 
        {
            int sommet = Thint(tThint,i); 
            if (lesTh[sommet] == -1)
            {
                R3 & P =*(K.vertices[i]);
                t = promenade(Th,P,t);
                lesTh[sommet] = t; 
                // si trouver je peux coloriser le sommet
                if(t!=-1) {Thint.vertices[sommet].colorise();}
 	        // A chaque fois que je suis ici je fais le calcul pour un point
 	        static int statu; statu++;
  	        if((statu%FREQU_AFFICHE)==0)
		{  nb_tetraedre_parcourus_total += nb_tetraedre_parcourus;
	           cout<<"La moyenne des tetraedres visites par promenande est de "<<double(nb_tetraedre_parcourus)/double(FREQU_AFFICHE)<<" pour "<<FREQU_AFFICHE<<" points."<<endl;
		   nb_tetraedre_parcourus = 0;
		} 		
            }
        }
        for (int i=0;i<4;i++) 
        {
            if ((K.Neigh[i] != -1) && (Thint.tetraedres[K.Neigh[i]].color ==0))
            {
                PThint.push(K.Neigh[i]);       
                Thint.tetraedres[PThint.top()].colorise();    
                if (t==-1)
                {  t=tt; }
                PTh.push(t);          
            }          
        }
    }    
    nb_tetraedre_parcourus_total += nb_tetraedre_parcourus;
    cout<<"La moyenne totale des tetraedres visites par promenande est de "<<double(nb_tetraedre_parcourus_total)/double(Thint.nv)<<endl;
}
/*
 * calcul les valeurs interpolees sur le second maillage
 * lesTh[i] contient le numero du tetraedre du premier maillage
 * contenant le point 'i' du maillage  a interpoler
 */
void interpole(Result &Re, Result &Reinter, Mesh &Thint, Mesh &Th, int *lesTh)
{
   // selon le type de solution   
   switch(Reinter.type)
   {
      default:
        cerr << "Type inconnu "<< Reinter.type << ", on ne sait pas traiter " << endl;
	throw(ErrorExec("exit",1)); 	  
	break;
      case RESULT_TYPE_SCALAIRE:
      {
         for(int i=0;i<Thint.nv;i++)
         {
           if(lesTh[i]!=-1)
           {  R3 & P = Thint(i);
              interpolation_scalaires(Reinter,Re,i,P,Th,lesTh[i]); }
         }
	 break;	 
      }
      case RESULT_TYPE_VECTORIEL:
      {
         for(int i=0;i<Thint.nv;i++)
         {
           if(lesTh[i]!=-1)
           {  R3 & P = Thint(i);
              interpolation_velocity(Reinter,Re,i,P,Th,lesTh[i]);}
         }
	 break;
      }
   };  

   cout<<endl<<endl;
   // On regarde si tout les points ont ete trouver
   int pointpastrouver = 0;    
   for(int i=0;i<Thint.nv;i++)
   {  if(lesTh[i]==-1) {pointpastrouver++;}}

   if(pointpastrouver ==0)
   {
      cout<<"Tous les points ont ete traites"<<endl;
   }
   else
   {

      cout<<"***Une autre interpolation va s'effectuer pour tous les points non traiter***"<<endl;
      cout<<"Il y a exactement "<<pointpastrouver<<" pour lesquels promenade n'a rien pu faire!!"<<endl;
      // Tout d'abord on commence par construire le chainage des points -> tetraedre
      int * Head;
      int * next;
      int * chaineTet;
      Head = new int [Thint.nv];
      next = new int [4*Thint.nt];
      chaineTet = new int [4*Thint.nt];
      for(int i=0;i<Thint.nv;i++)
      {  Head[i] = -1; }  //pas dans l'image      
      for(int i=0;i<4*Thint.nt;i++)
      {
            int s = Thint(i/4,i%4);
            next[i] = Head[s];
            chaineTet[i]=i/4;
            Head[s] = i;
      }
      
      // Stockage des points non fait
      int dd=0;
      for (int i=0;i<Thint.nv;i++)
	      if(lesTh[i]==-1) dd++;
      Pile Psommet(dd);
      Pile Psomtmp(dd);
      for (int i=0;i<Thint.nv;i++)
	      if(lesTh[i]==-1)
		      Psommet.push(i);

      int fautfinir = -1;
      while(fautfinir==-1)
      {   int kpile = Psommet.kpile();
          while(Psommet.kpile()>0 )
          {  int som = Psommet.pop();   
             int compte = 0;
             for(int p=Head[som];p!=-1;p=next[p])
		   switch(Reinter.type) {
			   case RESULT_TYPE_SCALAIRE: { compte += nouvelle_interpolation_scalaires(Reinter,Thint,som,chaineTet[p]); break; }
			   case RESULT_TYPE_VECTORIEL: { compte += nouvelle_interpolation_velocity(Reinter,Thint,som,chaineTet[p]); break; }
		   };  
             // si pas au moins un des tetraedres sinon en reempile dans l'autre liste
             // sinon on fait la moyenne des valeurs trouver et on colorise le point
             if(compte == 0)
             { Psomtmp.push(som);}
             else
             {
		   switch(Reinter.type) {
			   case RESULT_TYPE_SCALAIRE:
				   Reinter.scalaires[som] /= compte;  break;
			   case RESULT_TYPE_VECTORIEL:
				   Reinter.velocity[som] /= compte;  break;
		   };  
               Thint.vertices[som].colorise(); // on colorise le points car on pourra s'en reservir!
               lesTh[som] = Thint.nv + 1; // histoire de !!!
             } 
          }
          if( Psomtmp.kpile()==0)
          {  fautfinir = 1;}
          else
          {   // si cette condition est verifiee il faut sortir sinon boucle sans fin
              assert(Psomtmp.kpile()!=kpile);
//	      cout  << "Psomtmp.kpile() et kpile  : " << Psomtmp.kpile() << " " << kpile << endl;
              while(Psomtmp.kpile()>0)
                 Psommet.push( Psomtmp.pop());
          }
      }
   } 
}



void sauve_solution(Result & Reinter, const char *resultat)
{

   // selon le type de solution   
   switch(Reinter.type){
      default:
        cerr << "Type inconnu "<< Reinter.type << ", on ne sait pas traiter " << endl;
	throw(ErrorExec("exit",1)); 	  
	break;
      case RESULT_TYPE_SCALAIRE:
      {
         // sauvegarde du fichier
         ofstream fs(resultat);
         fs<<Reinter.dim<<" "<<Reinter.type<<" "<<Reinter.nv<<" "<<Reinter.typedonnees<<endl;
         for(int i=0;i<Reinter.nv;i++)
         { fs <<Reinter.scalaires[i]<<endl; }
         fs.close();
	 break;	 
      }
      case RESULT_TYPE_VECTORIEL:
      {
         // sauvegarde du fichier
         ofstream fv(resultat);
         fv<<Reinter.dim<<" "<<Reinter.type<<" "<<Reinter.nv<<" "<<Reinter.typedonnees<<endl;
         for(int i=0;i<Reinter.nv;i++)
         { fv <<Reinter.velocity[i]<<endl;  }
         fv.close();
	 break;
      }
   };
   cout<<"Le resultat vous sera donne dans le fichier \""<<resultat<<"\""<<endl;        
}

inline double CPUtime(){
#ifdef SYSTIMES
  struct tms buf;
  if (times(&buf)!=-1)
    return ((double)buf.tms_utime+(double)buf.tms_stime)/(long) sysconf(_SC_CLK_TCK);
  else
#endif
    return ((double) clock())/CLOCKS_PER_SEC;
}


// debut du programme
int main(int argc, char **argv)
{
	// d'abord on voit si on doit faire un test
	if ( argc>1 && ( !strcmp(argv[1], "--test")  || !strcmp( argv[1], "-t" ))) {
		const char * maillage1 = (argc<3) ? "cube2_3d.mesh" : argv[2]; 
		const char * maillage2 = (argc<4) ? "cube0_3d.mesh" : argv[3]; 
		test(maillage1, maillage2);
		return 0;

	}



	for ( int i=1; i<argc; i++ ) {
		if ( !strcmp(argv[i], "--version")  || !strcmp( argv[ i ], "-v" )) {
			printf(VERSION_STRING "\n");
			exit( 0 );
		}
		if ( !strcmp(argv[i], "--help")  || !strcmp( argv[ i ], "-h" )) {
			printf("Usage : %s [fichierresultat.bb [donnees.bb [maillage1.mesh [maillage2.msh ] ] ] ]\n", argv[0]);
			printf("      : --help or -h    : print this help message and exit\n", argv[0]);
			printf("      : --version or -v : print the version and exit\n", argv[0]);
			printf("Usage : %s --test [maillage1.mesh [maillage2.msh ] ]\n", argv[0]);

			printf("\n" );
			exit( 0 );
		}
	}

   cout<<endl<<endl;
   // Declarartion par default des differents maillages
   const char * resultat  = (argc<2) ? "resultat.bb"   : argv[1]; 
   const char * solution  = (argc<3) ? "cube2_3d.bb"   : argv[2];    
   const char * maillage1 = (argc<4) ? "cube2_3d.mesh" : argv[3]; 
   const char * maillage2 = (argc<5) ? "cube00_3d.mesh" : argv[4]; 


   // Construction du maillage ou l'on connait la solution
   
   R T0=CPUtime();
   Mesh Th(maillage1);  
   // Construction du maillage ou l'on doit faire l'interpolation de la solution
   Mesh Thint(maillage2);     
   // Le calcul du epsilon = hi pour Coord_bary
   R hi = Th.taille_caracteristique();
   R hiint = Thint.taille_caracteristique();
   hi = sqrt(hi);
   hiint = sqrt(hiint);
   cout << "Tailles caracteristiques : Th et Thint : " << hi << " " << hiint << endl;
   Epsilon *= Max(hi,hiint); 


#if 0
   // La boucle O(nv*nt)
   for(int i = 0;i<Thint.nv;i++)
   {
       R3 & P = Thint(i);
       for (int j=0;j<Th.nt;j++)
       {
           Tetraedre & K(Th[j]);
           R3 &A =*(K.vertices[0]);    R3 &B =*(K.vertices[1]);
           R3 &C =*(K.vertices[2]);    R3 &D =*(K.vertices[3]);
           if(coord_bary(P,A,B,C,D)>-Epsilon && coord_bary(P,B,C,D,A)>-Epsilon && coord_bary(P,C,D,A,B)>-Epsilon &&
           coord_bary(P,D,A,B,C)>-Epsilon)
           {
             lesTh[i] = j;
             Thint.vertices[i].colorise();
           }
       }
    }
#endif

#if 0
    // La boucle avec promenade O(nv * logn)
    // On part du tetraedre 0 car on ne sait pas ou se trouve the first tetraedre
    int t= 0; 
    int nb_tetraedre_parcourus_total = 0;
    for(int i = 0;i<Thint.nv;i++)
    {
        R3 & P = Thint(i);
        t = promenade(Th,P,t);
 	// A chaque fois que je suis ici je fais le calcul pour un point
 	static int statu; statu++;
  	if((statu%FREQU_AFFICHE)==0)
	{  nb_tetraedre_parcourus_total += nb_tetraedre_parcourus;
	   cout<<"La moyenne des tetraedres visites par promenande est de "<<double(nb_tetraedre_parcourus)/double(FREQU_AFFICHE)<<" pour "<<FREQU_AFFICHE<<" points."<<endl;
	   nb_tetraedre_parcourus = 0;
        } 	

        if(t !=-1)
        {  Thint.vertices[i].colorise(); lesTh[i]=t;}
        else
        {  t=0;}
    }   
    nb_tetraedre_parcourus_total += nb_tetraedre_parcourus;
    cout<<"La moyenne totale des tetraedres visites par promenande est de "<<double(nb_tetraedre_parcourus_total)/double(Thint.nv)<<endl;     
#endif

   // La solution que l'on connait
   Result Re(solution,Th);
   // La solution que l'on cherche
   Result Reinter(Thint,Re.type,Re.typedonnees);
   // le tableau ou on stock les tetraedres de Th correspondant au points de Thint
   int *lesTh  = new int[Thint.nv];
   
   R T1=CPUtime();
   cout<<endl<<"*******> Pour le chargement des donnees le temps ecoule est : "<<T1-T0<< " secondes."<<endl<<endl;   

   calcule_pile(Thint, Th, lesTh);
   interpole(Re, Reinter, Thint, Th, lesTh);
   sauve_solution(Reinter, resultat);
 
   R T2=CPUtime();   
   cout<<endl<<"*******> L'interpolation et la sauvegarde des donnees ont mis : "<<T2-T1<< " secondes."<<endl;      
   R T3=CPUtime();      
   cout<<endl<<"*******> Ce programme a mis en tout et pour tout : "<<T3-T0<< "
   secondes afin d'effectuer tous les calculs."<<endl;
 
   return 0;	
}



/*
 * Ce qui suit sert pour les tests
 */
R champs_scalaire_lineaire( R3 & P) 
{ 
   return 1. * P.x + 1. * P.y + 1. * P.z + 1.;
}  


void test(const char * maillage1,  const char *maillage2)
{
   R T0=CPUtime();
   // Construction du maillage ou l'on connait la solution
   Mesh Th(maillage1);  

   // Construction du maillage ou l'on doit faire l'interpolation de la solution
   Mesh Thint(maillage2);  

   // Le calcul du epsilon = hi pour Coord_bary
   R hi = Th.taille_caracteristique();
   R hiint = Thint.taille_caracteristique();
   hi = sqrt(hi);
   hiint = sqrt(hiint);
   cout << "Tailles caracteristiques : Th et Thint : " << hi << " " << hiint << endl;
   Epsilon *= Max(hi,hiint); 
   cout << "Epsilon vaut : " << Epsilon << endl;

   // La solution exacte sur le premier maillage
   Result Re(Th,1,2);
   for (int i=0;i<Re.nv;i++)
	   Re.scalaires[i] = champs_scalaire_lineaire(Th(i));
   
   // La solution exacte que l'on devrait trouver sur le second maillage
   Result Exacte(Thint,1,2);
   for (int i=0;i<Exacte.nv;i++)
	   Exacte.scalaires[i] = champs_scalaire_lineaire(Thint(i));

   // La solution que l'on cherche
   Result Reinter(Thint,Re.type,Re.typedonnees);

   int *lesTh = new int[Thint.nv];
   for(int i=0;i<Thint.nv;i++) lesTh[i] = -1;
   
   R T1=CPUtime();
   cout<<endl<<"*******> Pour le chargement des donnees le temps ecoule est : "<<T1-T0<< " secondes."<<endl<<endl;   


   calcule_pile(Thint, Th, lesTh);
   interpole(Re, Reinter, Thint, Th, lesTh);

   R T2=CPUtime();   
   cout<<endl<<"*******> L'interpolation et la sauvegarde des donnees a mis : "<<T2-T1<< " secondes."<<endl;      
      
   cout<<endl<<endl;     
   //calcul d'erreur
   R err_RN;
   err_RN=0;
   for (int i=0; i<Thint.nv; i++)
   {
      err_RN += (Reinter.scalaires[i]-Exacte.scalaires[i])*(Reinter.scalaires[i]-Exacte.scalaires[i]);
   }

   err_RN= sqrt(err_RN);
   err_RN /= Thint.nv;

   R err_Linf=0;
   for (int i=0; i<Thint.nv; i++)
   {
      err_Linf = max (err_Linf,abs(Reinter.scalaires[i]-Exacte.scalaires[i]));
   }

   R err_L2=0.;
   for (int i=0; i<Thint.nt; i++)
   {
      const Tetraedre & K(Thint[i]);
      R volume =K.volume;
      R s=0.;
      for (int k=0;k<4;k++)
      {
          int num = Thint.number(K.vertices[k]);
	  s = s + ((Reinter.scalaires[num]-Exacte.scalaires[num])*(Reinter.scalaires[num]-Exacte.scalaires[num]));
      }
      err_L2 = err_L2 + (volume * s)/4.;
   }
   err_L2 = sqrt(err_L2);
   err_L2 /= Thint.nt;

   cout <<"L'erreur en norme infinie est de       : "<<err_Linf<<endl;    
   cout <<"L'erreur en norme euclidienne est de   : "<<err_RN<<endl;
   cout <<"L'erreur en norme L2 est de            : "<<err_L2<<endl;

   R T3=CPUtime();      
   cout<<endl<<"*******> Ce programme a mis en tout et pour tout :"<<T3-T0<< " secondes affin d'effectuer tout les calculs."<<endl;
   cout<<endl<<endl;

}


