
// std c
#include <stdlib.h>

// Thomas
#include "particule.h"


#define RAYON_GROSSE_BOULE 3


void ParticuleSystem::creation_sur_cube(int nbparticules)
{
	nb = nbparticules;
	particules = new Particule[nbparticules];
	Q_ASSERT( particules );

	// On mets les boules sur un cube, en hauteur, entre
	// au milieu du tore, il aura donc tendance a imploser
	int size = int( pow(double(nb), 1./3.) );
	R x,y,z,dx;
	y = x = BOXSIZE/4;    //         |            x            |
	                      //         |    xxxxxxxxxxxxxxx      |
	z = BOXSIZE-1;
	dx = BOXSIZE/2;
	dx /= double( size +0.1);

	qDebug("creation_sur_cube : Creation de %d spheres : dx vaut %f", nb, dx);
	
	for ( int i=0; i<nb; i++ ) {
		particules[i].init( R3( x,y,z ) );
		x+=dx;
		if ( x> 3*BOXSIZE/4 ) {y+=dx; x = BOXSIZE/4; }
		if ( y> 3*BOXSIZE/4 ) {z-=dx; y = BOXSIZE/4; }
//		printf("creating sphere at %f,%f,%f\n",x,y,z );
	}
}

void ParticuleSystem::creation_aleatoire(int nbparticules)
{
	nb = nbparticules;
	particules = new Particule[nbparticules];
	qDebug("premiere particule : %p", &particules[0]);
	qDebug("derniere particule : %p", &particules[nb-1]);
	Q_ASSERT( particules );

	// au hasard..

	qDebug("creation_aleatoire : Creation de %d spheres", nb);
	
#define HAUTEUR (BOXSIZE/4)
	for ( int i=1; i<nb; i++ ) {
		int x,y,z;
		x = random() % BOXSIZE;
		y = random() % BOXSIZE;
		z = random() % HAUTEUR;
		z+= HAUTEUR_CAR-HAUTEUR/2;// z varie autour de HAUTEUR_CAR, sur une hauteur HAUTEUR

		particules[i].init( R3( x,y,z ) );
//		printf("creating sphere at %f,%f,%f\n",x,y,z );
	}
	// la particule 0 est speciale : grosse et rouge
	particules[0].init( R3( BOXSIZE/2,BOXSIZE/2,BOXSIZE-1 ) ); // on lache une particule
	particules[0].setColor(0);
	particules[0].masse = RAYON_GROSSE_BOULE*RAYON_GROSSE_BOULE*RAYON_GROSSE_BOULE;
	particules[0].rayon = RAYON_GROSSE_BOULE;
}



void ParticuleSystem::creation_quinconce(int nbparticules)
{
	nb = nbparticules;
	particules = new Particule[nbparticules];
	qDebug("premiere particule : %p", &particules[0]);
	qDebug("derniere particule : %p", &particules[nb-1]);
	Q_ASSERT( particules );

	// sinusoidalement..

	qDebug("creation_quinconce  : Creation de %d spheres", nb);
	
#define HAUTEUR (BOXSIZE/4)


	double tmp = float(BOXSIZE)*float(BOXSIZE)*float(HAUTEUR); // volume total
	qDebug("tmp = %f", tmp);
	tmp /= nbparticules; // volume par particule
	qDebug("tmp = %f", tmp);
	tmp = pow( tmp, 1./3. ); // dimension par direction d'espace
	qDebug("tmp = %f", tmp);
	int d = int (tmp);
	d++;

	qDebug("d = %d", d);

	int hauteur = int(HAUTEUR/tmp); // nbre de particules en hauteur;
	++hauteur *= DCAR*1.5; // hauteur pour ~equilibre

	int x,y,z;
	x=y=d; z = HAUTEUR_CAR-hauteur/2;

	int togglex=0; 
	int toggley=0; 

	for ( int i=1; i<nb; i++ ) {
		x += d;
		if ( x>= BOXSIZE ) {
			x = ( togglex )? d:d>>1;
			togglex=!togglex;
			y += d;
			if ( y>= BOXSIZE ) {
				y = ( toggley )? d:d>>1;
				toggley=!toggley;
				z+=DCAR*1.5;
				qDebug("Creating (%d,%d,%d)", x, y ,z);
			}
		}
		particules[i].init( R3( x,y,z ) );
//		printf("creating sphere at %f,%f,%f\n",x,y,z );
	}
	// la particule 0 est speciale : grosse et rouge
	particules[0].init( R3( BOXSIZE/2,BOXSIZE/2,HAUTEUR_CAR+hauteur/2 + (hauteur/2+3)*DCAR ) ); // on lache une particule
	particules[0].setColor(0);
	particules[0].masse = RAYON_GROSSE_BOULE*RAYON_GROSSE_BOULE*RAYON_GROSSE_BOULE;
	particules[0].rayon = RAYON_GROSSE_BOULE;
}




void ParticuleSystem::creation_uniforme(int nbparticules)
{
	nb = nbparticules;
	particules = new Particule[nbparticules];
	Q_ASSERT( particules );

	// On mets les boules sur un cube, en hauteur, entre
	// au milieu du tore, il aura donc tendance a imploser
	int size = int( pow(double(nb), 1./3.) );
	R x,y,z,dx;
	y = x = 0;             //         |           x             |
	                       //         |    xxxxxxxxxxxxxxx      |
	R dz = R(BOXSIZE)/R(10);
	z = dz;
	dz/=4;
	dx = BOXSIZE;
	dx /= double( size +0.1);

	qDebug("creation_uniforme : Creation de %d spheres : dx vaut %f", nb, dx);
	
	for ( int i=0; i<nb; i++ ) {
		particules[i].init( R3( x,y,z ) );
		x+=dx;
		if ( x>= BOXSIZE-2*dx ) {y+=dx; x = 0; }
		if ( y>= BOXSIZE-2*dx ) {z+=dz; y = 0; }
//		printf("creating sphere at %f,%f,%f\n",x,y,z );
	}
}

