
// stdlib
#include <stdio.h>
#include <unistd.h> // usleep
#include <math.h>

// Qt
#include <qtimer.h>
#include <qimage.h>

// Projet
#include "particule.h"

//#define NB_PARTICULES (8*8*8)
//#define NB_PARTICULES (30*30*30)
//#define NB_PARTICULES (20*20*20)
//#define NB_PARTICULES (15*15*15)
#define NB_PARTICULES (2000)
//#define NB_PARTICULES ( 40 )

#define DESSIN_PERIODE 200
#define MAJ_CACHE_PERIODE 100
//#define STATS_PERIODE 10000
#define STATS_PERIODE  1000


/**
  * Particule::
  */

// les couleurs qu'on utilisent pour dessiner les boules.. (rgb)
const int Particule::colorMapSize = 7;
const double Particule::colorMap[][3] = {
		{ 1., 0., 0.},
		{ 0., 1., 0.},
		{ 0., 0., 1.},
		{ 0., 1., 1.},
		{ 1., 0., 1.},
		{ 1., 1., 0.},
		{ 1., 1., 1.},
};


#if SCHEMA_LONG==4
double schema_coeffs[] = {55./24., -59/24., 37./24., -3./8.};
#else
# if SCHEMA_LONG==6
double schema_coeffs[] = {4277./1440., -7923./1440., 4991./720., -3649./720., 959./480., -95./288.};
# else
# error SCHEMA_LONG unknown
# endif
#endif



/**
  * gere le fait que c'est un tore pour avoir la vraie
  * 'distance' entre deux particules
  */
R3   Particule::operator-(Particule &P)const
{
	R3 d;
	R l;

	d = position-P.position;

	// x
	l = d.x * d.x;
	if ( l> R(BOXSIZE/2)*R(BOXSIZE/2) ) // x is bad
		if ( position.x< R(BOXSIZE/2) )
			d.x += R(BOXSIZE);
		else 
			d.x -= R(BOXSIZE);

	// y
	l = d.y * d.y;
	if ( l> R(BOXSIZE/2)*R(BOXSIZE/2) ) // y is bad
		if ( position.y< R(BOXSIZE/2) )
			d.y += R(BOXSIZE);
		else 
			d.y -= R(BOXSIZE);

	return d;
}

void Particule::avance(void)
{
	R oldz = position.z;

	R3 v,a; // == 0 ds le constructeur

	// calcul la derniere vitesse
	for ( int r=0; r<SCHEMA_LONG; r++ )
		a+= schema_coeffs[r]*accel[(tabpos+r)%SCHEMA_LONG];
	v = vitesse[tabpos] +  FROTTEMENT * DT * a; // nouvelle vitesse

	// on se decale vers la gauche (-1)
//	qDebug("tabpos %d", tabpos);
	tabpos += SCHEMA_LONG-1;
	tabpos %= SCHEMA_LONG;
//	qDebug("tabpos %d\n", tabpos);
	vitesse[tabpos] = v; // on utilise la vitesse qu'on vient de calculer pour
	                     // calculer l'acceleration
			     //
	// calcul de la position
	v = R3 ( RZERO,RZERO,RZERO );
	for ( int r=0; r<SCHEMA_LONG; r++ )
		v+= schema_coeffs[r]*vitesse[(tabpos+r)%SCHEMA_LONG];
	position += DT * v;

	accel[tabpos] = R3( RZERO,RZERO,RZERO ); // acceleration courante

	// coordonnees toriques : (pas de %, on est en double)
	while ( position.x<RZERO ) position.x += BOXSIZE;
	while ( position.y<RZERO ) position.y += BOXSIZE;
	while ( position.x>=BOXSIZE ) position.x -= BOXSIZE;
	while ( position.y>=BOXSIZE ) position.y -= BOXSIZE;

	if ( position.z>=ZFACTOR*BOXSIZE ) {
		// si on depasse le z maximum : on coupe l'energie
		position.z = ZFACTOR*BOXSIZE-1;
		for ( int i=0; i<SCHEMA_LONG; i++ )
			vitesse[i]= accel[i]= R3( RZERO,RZERO,RZERO );
	}
	if ( position.z<RZERO ) {
		qDebug("oops, une particule(%p) tombe en dessous de 0, on triche..(%f->%f)", this, position.z, oldz);
		/*
		qDebug("tabpos : %d, z-speed : %f", tabpos, vitesse[tabpos].z);
		qDebug("z-accel : %f", accel[(tabpos+1)%SCHEMA_LONG].z);
		qDebug("z-accel : %f", accel[(tabpos+2)%SCHEMA_LONG].z);
		qDebug("z-accel : %f", accel[(tabpos+3)%SCHEMA_LONG].z);
		qDebug("z-accel : %f", accel[(tabpos+4)%SCHEMA_LONG].z);
		qDebug("z-accel : %f", accel[(tabpos+5)%SCHEMA_LONG].z);
		*/
		position.z = oldz;
		vitesse[tabpos].z *= 0.;
	}
}


void Particule::init(R3 p , R3 v)
{
	position=p;

	Q_ASSERT(p.x>=0); Q_ASSERT(p.x<BOXSIZE);
	Q_ASSERT(p.y>=0); Q_ASSERT(p.y<BOXSIZE);
	Q_ASSERT(p.z>=0); Q_ASSERT(p.z<ZFACTOR*BOXSIZE);

	for ( int i=0; i<SCHEMA_LONG; i++ ) {
		vitesse[i]= v;
		accel[i]= R3( RZERO,RZERO,RZERO );
	}
	tabpos = SCHEMA_LONG-1;
}

double Particule::energie(void)
{
	double v2 = Norme2_2(vitesse[tabpos]);
	return masse* (
			0.5*v2 + // cinetique
			// potentielle gravitationnelle + potentiel repulsion sol
			G*(position.z-pow( fabs( position.z )/double( HAUTEUR_CAR ) ,ETA+1 )/(ETA+1))
			);
}


double ParticuleSystem::energie(void)
{
	double e = 0.;
	/// potentielle gravitationnelle + cinetique
	for ( int i=0; i<nb; i++ ) // particule courante
		e+= particules[i].energie();
	// potentielle d'interaction
	for ( int i=0; i<nb; i++ )
		for ( int j=i+1; j<nb; j++ ) {
			Particule &p1 = particules[i];
			Particule &p2 = particules[j];
			R3 d = p1-p2;
			double l = Norme2( d );
			if ( l > DINF ) continue; // rien dans ce cas la
			if ( l<EPSILON ) continue; // trop petit
			e += p1.masse * p2.masse * pow(l/double(DCAR),ALPHA+1)/(ALPHA+1) - pow(l/double(DCAR ),BETA+1)/(BETA+1);
		}
	return e;
}

/**
  * ParticuleSystem
  */

ParticuleSystem::ParticuleSystem(QWidget* parent, const char* name,int nbparticules )
    : QGLWidget( parent, name )
{

	// nombre par defaut
	if ( nbparticules< 0) nbparticules = NB_PARTICULES;

	// Opengl
	xRot = 285.; yRot = 0.; zRot = 10.;
//	xRot = yRot = zRot = 0.;
//	xRot = 0.; yRot = 0.; zRot = 0.;
//	scale = .10;	// default object scale, the smaller, the more far away it will look
	scale = .14 *20/BOXSIZE;


//	creation_sur_cube(nbparticules);
//	creation_uniforme(nbparticules);
//	creation_aleatoire(nbparticules);
	creation_quinconce(nbparticules);

	// Cache :
	taille_cache = (BOXSIZE+int(DINF)-1)/int(DINF);
	taille_cache_z = (ZFACTOR*BOXSIZE+int(DINF)-1)/int(DINF);
	qDebug("Creation de  %d=%dx%dx%d caches", taille_cache*taille_cache*taille_cache_z, taille_cache, taille_cache, taille_cache_z);
	m_cache = new cache_t[ taille_cache*taille_cache*taille_cache_z ];
	Q_ASSERT( m_cache );
	maj_cache();
	qDebug("Fin de creation des caches");

	time.start(); nb_interactions = 0;  //initialisation pour les stats

#ifdef PR_THREAD_ALL
	// creation des threads
	thread_nb = taille_cache*taille_cache*taille_cache_z;
	qDebug("Creation du pool de threads (%d threads)",thread_nb);
	thread_pool = new psThread[thread_nb](this);
	for ( int x = 0; x < taille_cache; x++ )
		for ( int y = 0; y < taille_cache; y++ )
			for ( int z = 0; z < taille_cache_z; z++ )
				thread_pool[(x*taille_cache+y)*taille_cache_z+z].setCache(x,y,z);
#endif


#ifndef PR_THREAD_POOL
	// and timer
	QTimer *t = new QTimer( this ); // pas besoin de le detruire, il se detruit tout seul
	connect(  t, SIGNAL( timeout() ), SLOT( avance_un_pas() ) );
	t->start( 0, FALSE );// c'est parti... (see Qt doc)
#endif

#ifdef PR_THREAD_POOL
	// creation du pool de threads
	thread_nb = 4;
	qDebug("Creation du pool de threads (%d threads)",thread_nb);
	thread_pool = new psThread[thread_nb](this);
	for ( int i = 0; i < thread_nb; i++ )
		thread_pool[i].start();
#endif


}

void ParticuleSystem::maj_cache(void)
{
//	qDebug("Maj des caches");
	for ( int i = 0; i< taille_cache*taille_cache*taille_cache_z; i++ )
		m_cache[i].clear();
	for ( int i=0; i<nb; i++ ) // particule courante
		cache( particules[i] )->append(&particules[i]);
#if 0
	for ( int i = 0; i< taille_cache*taille_cache*taille_cache_z; i++ )
		if (m_cache[i].size()) qDebug( "cache[%d] has %d particules", i, m_cache[i].size() );
#endif
}


unsigned int ParticuleSystem::trouve_voisinage( int x, int y, int z, cache_t *voisin[7] )
{
	int k = 0; // les points
	Q_ASSERT(x>=0); Q_ASSERT(y>=0); Q_ASSERT(z>=0);

	// premier plan
	voisin[ k++ ] = _cache( ( x )           , (y+1)%taille_cache, z );
	voisin[ k++ ] = _cache( (x+1)%taille_cache, ( y )           , z );
	voisin[ k++ ] = _cache( (x+1)%taille_cache, (y+1)%taille_cache, z );

	if ( z==taille_cache_z-1 ) return k; // tout en haut : rien d'autre a faire
	z++; // le plan du dessus
	voisin[ k++ ] = _cache( ( x )           , ( y )           , z );
	voisin[ k++ ] = _cache( ( x )           , (y+1)%taille_cache, z );
	voisin[ k++ ] = _cache( (x+1)%taille_cache, ( y )           , z );
	voisin[ k++ ] = _cache( (x+1)%taille_cache, (y+1)%taille_cache, z );

	return k;
}

void ParticuleSystem::calcul_pour_un_cache(int x, int y, int z)
{
	cache_t *c =  _cache(x,y,z);
	Q_ASSERT(c);
	Particule *p;

	if ( !c->size() ) return; // rien a faire



	// forces d'interactions a l'interieur de "c"
	// localement O(n^2)

	cache_t::iterator it, it2;
        for (  it = c->begin(); it != c->end(); ++it ){
		p = *it;
        	for (  it2 = it+1; it2 != c->end(); ++it2 )
			calcule_interaction( *it2, p );
	}

	// force d'interaction : les voisins
	cache_t *vois[7];
	unsigned int nb = trouve_voisinage(x,y,z,vois);
       	for (  it2 = c->begin(); it2 != c->end(); ++it2 ){
		p = *it2;
		for ( unsigned int i = 0; i<nb; i++ ) {// boucle sur les caches
			cache_t::iterator e = vois[i]->end();
        		for (  it = vois[i]->begin(); it != e ; ++it )
				calcule_interaction(p, *it );
		}
	}


	// Repulsion du sol et force de pesanteur
       	for (  it2 = c->begin(); it2 != c->end(); ++it2 ){
		p = *it2;
		double dd = G * pow(fabs(p->position.z)/double(HAUTEUR_CAR) ,ETA );
//		qDebug( "Force exercee par le sol : %f", dd);
		p->ajoute_accel(R3(0,0,dd-G));
	}
}

void ParticuleSystem::calcule_interaction(Particule *P1, Particule *P2 )
{
//	if ( P1==P2 ) return;

	R3 d = *P1-*P2;
	double l = Norme2( d );
//	qDebug("distance is %f, DINF is %f", l, DINF);
	if ( l > DINF ) return; // rien dans ce cas la
	if ( l<EPSILON ) {
		qDebug("distance nulle.. on triche");
		return;
	}
	d *= force(l)/l; //  d/l est le vecteur unitaire
	                 //  d*force(l)/l est le vecteur force
	P1->ajoute_accel( d * P2->masse);
	P2->ajoute_accel(-d * P1->masse);
	nb_interactions++;
}


#ifdef PR_NO_THREAD
void ParticuleSystem::calcul_un_pas(void)
{
	// version classique sans thread
	for ( int x = 0; x < taille_cache; x++ )
		for ( int y = 0; y < taille_cache; y++ )
			for ( int z = 0; z < taille_cache_z; z++ )
				calcul_pour_un_cache(x,y,z);
}
#endif
#ifdef PR_THREAD_ALL
void ParticuleSystem::calcul_un_pas(void)
{

	// version avec pleins de threads / grossier
	qDebug("lance les threads");
	for (int i =0; i<thread_nb; i++) thread_pool[i].start(QThread::LowPriority);
//	qDebug("attends les threads");
	for (int i =0; i<thread_nb; i++) thread_pool[i].wait();
//	qDebug("fin des threads");
//	qDebug("");
}
#endif


#ifdef PR_THREAD_POOL
void ParticuleSystem::get_job(int &_x, int &_y)
{
	static int x=0,y=0;
	static QMutex wmutex;
	static QMutex calcul;
	static int waiting = 0; // waiting contient le nombre de threads qui sont ds get_job();

	wmutex.lock(); waiting++; wmutex.unlock();

	calcul.lock(); // lock toute les calculs sur les indices...

	_x = x; _y = y; // on donne le travail
	// incremente (x,y)
	if (++x>=taille_cache) {
		x=0;
		if (++y>=taille_cache) {
			// on a fait un tour ->
			while ( waiting !=  thread_nb ) {
				usleep(1); // synchronisation
				//qDebug("waiting");
			}
			avance_un_pas();
			y=0;
		}
	}
	calcul.unlock();

	wmutex.lock(); waiting--; wmutex.unlock();
}
#endif



void ParticuleSystem::avance_un_pas(void)
{

	// pour POOL, calcul_un_pas n'est pas utilise
#ifndef PR_THREAD_POOL
	calcul_un_pas();
#endif
//	qDebug("avance_un_pas()");

	static int etapes_per=STATS_PERIODE-2;
	static int etapes=0;
	etapes++;
	if ( ++etapes_per> STATS_PERIODE ) {
		qDebug("Quelques stats : ");
		qDebug("    Debit en k-interaction/s : %f", float(nb_interactions)/float(time.elapsed()) );
		qDebug("    Etapes de temps(total)   : %f k", double(etapes/1000.) );
		qDebug("    Etapes de temps/s        : %f", float(1000*etapes_per)/float(time.elapsed()) );
		qDebug("    Interactions/etapes      : %f", float( nb_interactions )/float( etapes ) );
		qDebug("    Interactions/etapes/part : %f", float( nb_interactions )/float( etapes )/float(NB_PARTICULES) );
#if 0
		double e = energie();
		if ( e>1000*2000 )
			qDebug("    Energie du systeme       : %f Mega", energie()/1000000. );
		else if ( e>2000 )
			qDebug("    Energie du systeme       : %f Kilo", energie()/1000. );
		else    qDebug("    Energie du systeme       : %f", energie() );
#endif
		int j=0;
		for ( int i = 0; i< taille_cache*taille_cache*taille_cache_z; i++ )
			if (m_cache[i].size()) j++;
		qDebug("    Caches non vides         : %d", j);


		nb_interactions = 0;
		time.start();
		etapes_per=0;
	}

	// tout le monde avance
	for ( int i=0; i<nb; i++ )
		particules[i].avance();

	// Mise a jour de la fenetre opengl
	static int draw_call=0;
	if (++draw_call> DESSIN_PERIODE ) {
		draw_call = 0;
		updateGL();
// Si on veut faire une animation :
		static int counter=0; grabFrameBuffer().save(QString().sprintf("MoleculesAnimation%06d.png",counter++ ),"PNG" );;
	}

	// Mise a jour du cache
	static int cache_call=0;
	if (++cache_call> MAJ_CACHE_PERIODE ) {
		cache_call = 0;
		maj_cache();
	}

	// lorsque on veut faire du gprof, il faut quitter proprement -> hop
#if 0
	static int toto=0;
	if ( ++toto > 3000 ) exit(0);
#endif

}

