#ifndef PARTICULE_H
#define PARTICULE_H

// C
#include <math.h>

// C++
#include <iostream>  //streams

// Qt
#include <qgl.h>
#include <qdatetime.h>
#include <qvaluevector.h> 
#include <qthread.h> 

#define BOXSIZE  ( 10000 )
#define ALPHA    ( 2.*BETA)           // force de repulsion quand tres proche
#define BETA     ( -2. )              // force d'attraction quand pas trop proche
#define DCAR     ( BOXSIZE/60 )       // distance caracteristique, qui correspond a
                                      //    l'invertion attraction/repulsion
#define HAUTEUR_CAR (BOXSIZE/4)       // hauteur a laquelle la repulsion du sol == gravitation
#define RADIUS   ( DCAR/2. )          // Rayons des boules
#define FROTTEMENT ( 0.9 )            // coefficient de frottement, 1.0 == pas de frottement
                                      //       0.0 == ca avance plus
#define EPSILON  (double(DCAR)/1000.) // Distance a partir de laquelle on considere
                                      //    que deux boules sont trop proches
#define ZFACTOR  1                    // rapport entre la hauteur qu'on etudie, et la
                                      //    taille de la base utilisee pour le maillage
#define DINF     (5*DCAR)             // distance a partir de laquelle on neglige l'interaction
#define ETA      ( BETA)              // repulsion du sol
#define DT       (.03)
#define G        ( 0.49)              // force constante vers le bas (gravitation)

// decommenter un seul des suivants
#define PR_NO_THREAD
//#define PR_THREAD_POOL
//#define PR_THREAD_ALL

using namespace std; //iostreams..


// definition R  ---------------------------------------------------------------------------
typedef double R;
#define RZERO (0.)

// The class R3  ---------------------------------------------------------------------------
class R3 {
	friend ostream& operator <<( ostream& f, const R3 & P )
	{ f << P.x << ' ' << P.y << ' ' << P.z   ; return f; }
	friend istream& operator >>( istream& f,  R3 & P )
	{ f >> P.x >>  P.y >>  P.z  ; return f; }

	public:  
	R  x,y,z;

	R3 () :x(RZERO),y(RZERO),z(RZERO) {};
	R3 (R a,R b,R c):x(a),y(b),z(c)  {}
	void set (R a,R b,R c){x=a;y=b;z=c;}
	R3 (const R3 & A,const R3 & B) :x(B.x-A.x ),y( B.y-A.y),z(B.z-A.z)  {}
	R3 & operator=(R3 P2) {x=P2.x;y=P2.y;z=P2.z;return *this;}
	R3   operator+(R3 P)const   {return R3(x+P.x,y+P.y,z+P.z);}
	R3 & operator+=(R3 P)  {x += P.x;y += P.y;z += P.z;return *this;}
	R3   operator-(R3 P)const   {return R3(x-P.x,y-P.y,z-P.z);}
	R3 & operator-=(R3 P) {x -= P.x;y -= P.y;z -= P.z;return *this;}
	R3   operator-()const  {return R3(-x,-y,-z);}
	R3   operator+()const  {return *this;}
	R    operator,(R3 P)const  {return  x*P.x+y*P.y+z*P.z;} //produit scalaire
	R3  operator^(R3 P)const {return R3(y*P.z-z*P.y ,P.x*z-x*P.z, x*P.y-y*P.x);} //produit vectoriel
	R3   operator*(R c)const {return R3(x*c,y*c,z*c);}
	R3   operator/(R c)const {return R3(x/c,y/c,z/c);}
	R3   operator*=(R c)  {x *= c;y *= c; z *=c; return *this;}
	R3   operator/=(R c)  {x /= c;y /= c; z /=c; return *this;}
	//  R3   operator/=(double d)  {x /= d;y /= d; z /=d; return *this;}
	R  &  operator[](int i){ return (&x)[i];}
	friend R3 operator*(R c,R3 P) {return P*c;}
	friend R3 operator/(R c,R3 P) {return P/c;}
};

inline double Norme2_2(const R3 & A){ return double((A,A));} // norme au carre (bcp plus rapide)
inline double Norme2(const R3 & A){ return sqrt(double((A,A)));} // norme 

/**
  * quel schema ?? on a le choix entre 4 et 6 pour l'instant..
  */
#define SCHEMA_LONG 4

class Particule {

public:
	Particule (R3 p = R3(RZERO,RZERO,RZERO), R3 v = R3(RZERO,RZERO,RZERO)) {
		init( p,v );
		color=1+random()%(colorMapSize-1);// everything but red
		rayon = masse = 1.;
	}

	void init (R3 p , R3 v = R3(RZERO,RZERO,RZERO));
	void setColor( int c ) { color = c;}
	void dessiner(void);
#ifdef PR_NO_THREAD
	void ajoute_accel( R3 a ) { accel[tabpos] += a; }
#else
	void ajoute_accel( R3 a ) { mutex.lock(); accel[tabpos] += a; mutex.unlock(); }
#endif
	void avance(void);

	R3 position;
	double masse;
	double rayon;

	double energie( void );

	R3   operator-(Particule &P)const;

	// peut etre utile pour le debugage
	friend ostream& operator <<( ostream& f, const Particule & P )
	{ f<< "p= "<< P.position <<" v= "<<P.vitesse[P.tabpos] << " a=" << P.accel[P.tabpos] ; return f; }

protected:
	int color;
	R3 vitesse[ SCHEMA_LONG ];
	R3 accel[ SCHEMA_LONG ];
	int tabpos; // position dans le tableau, modulo SCHEMA_LONG

	static const int colorMapSize;
	static const double colorMap[][3];

#ifndef PR_NO_THREAD
private:
	QMutex mutex;
#endif
};

class ParticuleSystem : public QGLWidget
{
	friend class psThread;
	Q_OBJECT
public:
	ParticuleSystem(QWidget* parent, const char* name,int nbParticules=-1 );
	virtual ~ParticuleSystem();

	double energie( void );

private slots:
	void avance_un_pas(void);  // appele par la boucle d'evenement des qu'il 
                                   //    y a du temps disponible

	// ca, c'est pour le rendu opengl, c'est appele par
	// l'interface graphique.
public slots:
	void		setXRotation( int degrees );
	void		setYRotation( int degrees );
	void		setZRotation( int degrees );

private:
	Particule *particules;   // le tableau (dynamique) de particule
	int nb;                  // le nombre de particule de ce systeme

	/**
	  * rendu opengl
	  * voir rendu.cpp
	  */
	GLfloat xRot, yRot, zRot, scale;
	void		initializeGL();
	void		paintGL();
	void		resizeGL( int w, int h );


	/**
	  * gestion du cache
	  */
	typedef QValueVector<Particule*> cache_t;
	cache_t  *m_cache;
	int taille_cache;     // nbre de cache selon l'axe x ou y
	int taille_cache_z;   // nbre de cache selon l'axe z
	/**
	  * construction des caches
	  */
	void maj_cache(void);
	/*
	 * acces au tableau, fait l'equivalent de cache[x,y,z]
	 * les arguments = n. de cache ds chacune des directions
	 */
	cache_t  * _cache( int x, int y, int z ) {
		Q_ASSERT( x>=0 ); Q_ASSERT( x<taille_cache );
		Q_ASSERT( y>=0 ); Q_ASSERT( y<taille_cache );
		Q_ASSERT( z>=0 ); Q_ASSERT( z<taille_cache_z );
		Q_ASSERT( (x*taille_cache+y)*taille_cache_z+z<taille_cache*taille_cache*taille_cache_z );
		return & m_cache[(x*taille_cache+y)*taille_cache_z+z ]; }
	/*
	 * retourne le cache correspondant
	 * les arguments = coordonnes dans l'espace...
	 */
	cache_t  * cache( double x, double y, double z ) {
		return _cache( int(x/DINF), int(y/DINF), int(z/DINF)); }
	/*
	 * retourne le cache correspondant a la particule donnee
	 */
	cache_t  * cache( Particule &p ) { return cache( p.position.x, p.position.y, p.position.z );}

	/*
	 * renvoie dans voisins les adressses  des cache voisins pour le cache(x,y,z)
	 * renvoie le nombre de voisins, pour chaque axe dans une seule direction
	 * soit, 7 la plupart du temps ( 2^d-1 ) avec d = 3
	 * ou    3 si la cellule est en haut.. ( 2^d-1 ) d=2, on laisse tomber la 3eme D
	 */
	unsigned int trouve_voisinage( int x, int y, int z, cache_t *voisins[7] );
	/*
	 * le gros du travail se fait la.
	 */
	void      calcule_interaction(Particule *, Particule * );
	/**
	  * calcule pour tous les points contenus dans ce cache
	  */
	void            calcul_pour_un_cache(int x, int y, int z);
	/**
	  * norme de la force en fonction de la distance entre 2 particules
	  * TODO : a optimiser avec un tableau pres-calcule ???
	  */
	double force( double distance ) { return  ( pow(distance/double(DCAR),ALPHA) - pow(distance/double(DCAR ),BETA));}

	/**
	  * pour positionner les boules au debut
	  * -> ds le fichier creation.cpp
	  */
	void creation_sur_cube(int nbparticules);
	void creation_uniforme(int nbparticules);
	void creation_aleatoire(int nbparticules);
	void creation_quinconce(int nbparticules);

	int nb_interactions;// pour les stats
	QTime  time;        // pour les stats
#ifndef PR_NO_THREAD
	psThread           *thread_pool;
	int	            thread_nb;
#endif

#ifdef PR_THREAD_POOL
	void    get_job(int &_x, int &_y);
#else
	void calcul_un_pas(void);  // fait les lourds calculs, eventuellement avec des threads
#endif
};


#ifndef PR_NO_THREAD

#  ifndef QT_THREAD_SUPPORT
#  error QT_THREAD_SUPPORT must be defined
#  endif

class psThread : public QThread 
{
public:
	psThread(ParticuleSystem *p):QThread() { parent = p;}
	void setCache(int a, int b, int c) {x=a;y=b;z=c;}
#  ifdef PR_THREAD_POOL
	virtual void run(void) {
		for ( ;; ) {
			parent->get_job(x,y);
			for ( z=0; z<parent->taille_cache_z; z++ )
				parent->calcul_pour_un_cache(x,y,z);
		}
	}
#  else
	virtual void run(void) { parent->calcul_pour_un_cache(x,y,z); }
#  endif
private:
	int x,y,z;
	ParticuleSystem *parent;
};
#endif // no thread


#endif  //PARTICULE_H
