
// OpenGL
#include <GL/glu.h>

// Projet
#include "particule.h"



#define SPHERE 1


void Particule::dessiner(void )
{

	GLUquadricObj* mySphere = gluNewQuadric(); 
	glPushMatrix(); 
//		printf("drawing sphere at %f,%f,%f\n", particules[i].position.x, particules[i].position.y, particules[i].position.z);
	glTranslatef(	double(position.x ),
			double(position.y ),
			double(position.z ));

//	glNewList( SPHERE, GL_COMPILE );
		glColor3f( colorMap[color][0], colorMap[color][1], colorMap[color][2] );
		gluSphere( mySphere, RADIUS*rayon, 10, 10 ); 
//	glEndList();
	glPopMatrix(); 

}




void ParticuleSystem::paintGL()
{

	// on tourne..
	zRot+= 1.;
	if ( zRot>=360. ) zRot-=360.;

	// init 
	glClear( GL_COLOR_BUFFER_BIT );
	glLoadIdentity();

	glShadeModel( GL_SMOOTH );

	// lightening
	glEnable(GL_LIGHTING);
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
//	glEnable(GL_NORMALIZE);

	static GLfloat LightDiffuse[]= { 1.0f, 1.0f, 1.0f, 1.0f }; 
//	static GLfloat LightAmbient[]= { 1.0f, 1.0f, 1.0f, 1.0f }; 
//	static GLfloat LightAmbient[]= { 0.5f, 0.5f, 0.5f, 1.0f };
	static GLfloat LightAmbient[]= { 0.05f, 0.05f, 0.05f, 1.0f };
	static GLfloat LightPosition0[] = { BOXSIZE*7.0, BOXSIZE*5.0, BOXSIZE*16.0, 0.0 };
	static GLfloat LightPosition1[] = { -BOXSIZE*4.0, BOXSIZE*5.0, -BOXSIZE*6.0, 0.0 };

//	glLightfv( GL_LIGHT0, GL_AMBIENT, LightAmbient );
	glLightfv( GL_LIGHT0, GL_DIFFUSE, LightDiffuse );
	glLightfv( GL_LIGHT0, GL_SPECULAR,LightDiffuse );
	glLightfv( GL_LIGHT0, GL_POSITION,LightPosition0 );
	glEnable(GL_LIGHT0);

//	glLightfv( GL_LIGHT1, GL_AMBIENT, LightAmbient );
	glLightfv( GL_LIGHT1, GL_DIFFUSE, LightDiffuse );
	glLightfv( GL_LIGHT1, GL_SPECULAR,LightDiffuse );
	glLightfv( GL_LIGHT1, GL_POSITION,LightPosition1 );
	glEnable(GL_LIGHT1);



	// Remember, last transform call is performed _first_
	glTranslatef( .0, .0 , -10. ); // end : put the Camera
	glScalef( scale, scale, scale ); // zooming
	glRotatef( xRot, 1.0, 0.0, 0.0 ); 
	glRotatef( yRot, 0.0, 1.0, 0.0 ); 
	glRotatef( zRot, 0.0, 0.0, 1.0 );

//	glTranslatef( -.5, -.5 , -.5 );

    // first center the view
	glTranslatef(   -double(BOXSIZE)/2.0,
			-double(BOXSIZE)/2.0,
			-double(BOXSIZE)/2.0 );

	for ( int i=0; i<nb; i++ )
		particules[i].dessiner();

}

/*!
  Set up the OpenGL rendering state, and define display list
*/

void ParticuleSystem::initializeGL()
{
    qglClearColor( black ); 		// Let OpenGL clear to black
    glShadeModel( GL_FLAT );
}



/*!
  Set up the OpenGL view port, matrix mode, etc.
*/

void ParticuleSystem::resizeGL( int w, int h )
{
    glViewport( 0, 0, (GLint)w, (GLint)h );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glFrustum( -1.0, 1.0, -1.0, 1.0, 5.0, 15.0 );
    glMatrixMode( GL_MODELVIEW );
}


/*!
  Set the rotation angle of the object to \e degrees around the X axis.
*/

void ParticuleSystem::setXRotation( int degrees )
{
    xRot = (GLfloat)(degrees % 360);
    updateGL();
}


/*!
  Set the rotation angle of the object to \e degrees around the Y axis.
*/

void ParticuleSystem::setYRotation( int degrees )
{
    yRot = (GLfloat)(degrees % 360);
    updateGL();
}


/*!
  Set the rotation angle of the object to \e degrees around the Z axis.
*/

void ParticuleSystem::setZRotation( int degrees )
{
    zRot = (GLfloat)(degrees % 360);
    updateGL();
}


/**
  * destructeur
  */
ParticuleSystem::~ParticuleSystem()
{
	makeCurrent();

	// timer : pas besoin
	delete [] m_cache;
	delete [] particules;
#ifndef PR_NO_THREAD
	for ( int i=0; i<thread_nb;i++ )
		thread_pool[i].terminate();
	for ( int i=0; i<thread_nb;i++ )
		thread_pool[i].wait();
	delete [] thread_pool;
#endif
}

